/*
   SPDX-FileCopyrightText: 2008 Michael Jansen <kde@michael-jansen.biz>

   SPDX-License-Identifier: LGPL-2.0-or-later
*/
#ifndef LIBKHOTKEYSFWD_H
#define LIBKHOTKEYSFWD_H

namespace KHotKeys
{
class Action;
class ActionDataBase;
class ActionDataGroup;
class ActionList;
class CommandUrlAction;
class DBusAction;
class MenuEntryAction;
class MenuEntryShortcutActionData;
class ShortcutTrigger;
class SimpleActionData;
class Trigger;
}

#endif /* #ifndef LIBKHOTKEYSFWD_H */
